/*************************************************************************

  This file is part of the ParaNut project.

  Copyright (C) 2022-2023 Elias Schuler <elias.schuler@hs-augsburg.de>
                2022-2023 Lukas Bauer <lukas.bauer@hs-ausgsburg.de>
      Hochschule Augsburg, University of Applied Sciences

  Redistribution and use in source and binary forms, with or without modification,
  are permitted provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice, this
     list of conditions and the following disclaimer.

  2. Redistributions in binary form must reproduce the above copyright notice,
     this list of conditions and the following disclaimer in the documentation and/or
     other materials provided with the distribution.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
  ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 **************************************************************************/

#include "slib_fifo.h"

#include <stdio.h>
#include <signal.h>
#include <systemc.h>

#define CLK_PERIOD 10.0

struct Tb : sc_module
{
    sc_in_clk clk {"clk"};
    sc_signal<bool> rst {"rst"};
    sc_signal<bool> clear {"clear"};
    sc_signal<bool> write {"write"};
    sc_signal<bool> read {"read"};
    sc_signal<sc_uint<SLIB_FIFO_WIDTH> > d {"d"};
    sc_signal<sc_uint<SLIB_FIFO_WIDTH> > q {"q"};
    sc_signal<bool> empty {"empty"};
    sc_signal<bool> full {"full"};
    sc_signal<sc_uint<SLIB_FIFO_SIZE_E> > usage {"usage"};



    SlibFifo slib_fifo {"slib_fifo"};


    SC_CTOR(Tb)
    {
        slib_fifo.clk_i (clk);
        slib_fifo.rst_i (rst);
        slib_fifo.clear_i (clear);
        slib_fifo.write_i (write);
        slib_fifo.read_i (read);
        slib_fifo.d_i (d);
        slib_fifo.q_o (q);
        slib_fifo.empty_o (empty);
        slib_fifo.full_o (full);
        slib_fifo.usage_o (usage);

        SC_CTHREAD(test_proc, clk.pos());
    }

    void test_proc(){


        #ifndef __SYNTHESIS__
        // Set cfg_debug_mode to suppress some simulation warnings
        pn_cfg_debug_mode = true;
        // Trace file ...
        sc_trace_file *tf;
        if (pn_cfg_vcd_level > 0) {
            tf = sc_create_vcd_trace_file ("slib_fifo_tb");
            tf->delta_cycles (false);

            //PN_TRACE
            PN_TRACE(tf, clk);
            PN_TRACE(tf, rst);
            PN_TRACE(tf, clear);
            PN_TRACE(tf, write);
            PN_TRACE(tf, read);
            PN_TRACE(tf, d);
            PN_TRACE(tf, q);
            PN_TRACE(tf, empty);
            PN_TRACE(tf, full);
            PN_TRACE(tf, usage);


            slib_fifo.Trace (tf, pn_cfg_vcd_level);
        } else {
            fprintf (stderr, "Tracing is disabled.\n");
            tf = NULL;
        }

        PN_INFO("Reseting...");
        rst = 1;
        wait(10);
        rst = 0;
        wait();

        while(true){
            PN_INFO("Running");
            PN_INFOF(("q was:\t 0%03x", (uint16_t) q.read()));
            PN_ASSERTM (q.read() == 0, "q is not zero.");
            write = 1;
            for(uint16_t i= 1; i < 70; i++){
                d = i;
                wait(1);
            };
            PN_INFO("Writing for 66 clock cycles to check for out of bound");
            PN_INFOF(("q was:\t 0x%03x", (uint16_t) q.read()));
            PN_ASSERTM ((uint16_t) q.read() == 0x001, "q is not 0x001.");
            write = 0;
            read = 1;
            wait(2);
            for(uint16_t i = 0; i < 70; i++){
                PN_INFOF(("Reading addr: \t %d", i));
                PN_INFOF(("q was:\t 0x%03x", (uint16_t) q.read()));
                if(i >= 0x40){
                    PN_ASSERTM ((uint16_t) q.read() == 0x01, "q is not 0x001.");
                    PN_ASSERTM (empty.read() == 1, "empty flag is not set");
                    break;
                } else {
                    PN_ASSERTM ((uint16_t) q.read() == i+1, "q is not addr.");
                }
                wait(1);
            }
            read = 0;
            d = 2;
            write = 1;
            wait(20);
            clear = 1;

            PN_INFO("Clearing in the middle of an Write, expecting the stuff written in d");
            PN_INFOF(("q was:\t 0x%03x", (uint16_t) q.read()));
            PN_ASSERTM ((uint16_t) q.read() == 0x002, "q is not 0x02");

            clear = 0;
            wait(2);
            wait(32);
            PN_INFO("Clear set low again and writing for 32 cycles");
            PN_INFOF(("q was:\t 0x%03x", (uint16_t) q.read()));
            PN_ASSERTM ((uint16_t) q.read() == 0x002, "q is not 0x02.");

            d = 4;
            read = 1;
            write = 1;
            wait(32);
            PN_INFO("Reading and Writing at the same time");
            PN_INFOF(("q was:\t 0x%03x", (uint16_t) q.read()));
            PN_ASSERTM ((uint16_t) q.read() == 0x002, "q is not 0x02.");

            rst = 1;
            wait(5);

            PN_INFO("Reset");
            PN_INFOF(("q was:\t 0x%03x", (uint16_t) q.read()));
            PN_ASSERTM ((uint16_t) q.read() == 0x002, "q is not 0x02.");

            rst = 0;
            read = 1;
            write = 0;
            wait(32);
            PN_INFO("Reading for 32 cycles after reset");
            PN_INFOF(("q was:\t 0x%03x", (uint16_t) q.read()));
            wait(200);
            PN_ASSERTM ((uint16_t) q.read() == 0x004, "q is not 0x04.");


            PN_INFO ("Simulation finished.");
            sc_stop();
            wait();
            if (tf) sc_close_vcd_trace_file (tf);
        }
        #endif

    }


};

// **************** Main ************************

int sc_main (int argc, char *argv[]) {
    int arg, cfg_help = 0;

    // Parse command line...
    arg = 1;
    while (arg < argc && argv[arg][0] == '-') {
        switch (argv[arg][1]) {
        case 't':
            pn_cfg_vcd_level = MAX (0, MIN (9, argv[arg][2] - '0'));
            fprintf (stderr, "(cfg) vcdLevel = %i\n", pn_cfg_vcd_level);
            break;
        case 'h':
            cfg_help = 1;
            break;
        default:
            printf ("PN_ERROR: Unknown option '%s'.\n", argv[arg]);
            arg = argc;
        }
        arg++;
    }
    if (cfg_help) {
        puts ("Usage: dm_tb [<options>]\n"
              "\n"
              "Options:\n"
              "  -t<n>: set VCD trace level (0 = no trace file; default = 2)\n");
        return 3;
    }


    sc_clock clk{"clk", sc_time(CLK_PERIOD, SC_NS)};
    Tb tb("tb");
    tb.clk(clk);
    sc_start ();
    cout <<"\n\t\t*****Simulation complete*****" << endl;


    return 0;
}